from django.contrib.contenttypes.models import ContentType
from django.forms import ModelChoiceField, ModelForm

from .models import FieldSanitizer


class ContentTypeChoiceFieldWithAppLabel(ModelChoiceField):
    """A ModelChoiceField for ContentTypes that displays the app_label in
    addition to the name.
    """
    def label_from_instance(self, obj):
        return u"{0}/{1}".format(obj.app_label, obj.name)


class FieldSanitizerAdminForm(ModelForm):
    content_type = ContentTypeChoiceFieldWithAppLabel(
        ContentType.objects.all().order_by('app_label', 'model')
    )

    class Meta:
        model = FieldSanitizer
        fields = [
            'content_type',
            'field_name',
            'tags',
            'attributes',
            'styles',
            'strip',
            'strip_comments',
        ]



"""
Management utility to clean Models that are registered with a FieldSanitizer.

"""
from django.core.management.base import BaseCommand
from janitor.models import FieldSanitizer, _clean_class_objects


class Command(BaseCommand):
    help = "Cleans all Models that have a related FieldSanitizer"

    def handle(self, *args, **options):
        sanitizers = FieldSanitizer.objects.all()
        klass_list = [fs.content_type.model_class() for fs in sanitizers]
        updated = _clean_class_objects(klass_list)
        self.stdout.write("\nCleaned {0} Objects\n".format(updated))



"""
Management utility to list HTML elements used in Existing Content.
This could be useful when trying to decide which elements to include
in the whitelist.
"""
from django.core.management.base import BaseCommand
from janitor.models import _get_tags_used_in_content


class Command(BaseCommand):
    help = (
        "Lists HTML elements used in all Models that have "
        "a related FieldSanitizer"
    )

    def handle(self, *args, **options):
        tag_list = _get_tags_used_in_content()
        tag_list = ", ".join(sorted(tag_list))
        msg = "\nTags used in content: {0}\n"
        self.stdout.write(msg.format(tag_list))



from html.parser import HTMLParser

from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction, IntegrityError
from django.db.models.signals import pre_save

from bleach import clean
from janitor import whitelists


def _register(callback, content_type_list):
    """Connects the `callback` function to each content type in the
    `content_type_list` with the `pre_save` signal handler."""
    for ct in content_type_list:
        pre_save.connect(
            callback,
            sender=ct.model_class(),
            dispatch_uid=ct.model
        )


def _j(some_list):
    """Shortcut for ``', '.join(some_list)``."""
    return ', '.join(some_list)


class FieldSanitizer(models.Model):
    content_type = models.ForeignKey(ContentType)
    field_name = models.CharField(
        max_length=255,
        help_text="The name of a field in the selected Model. It probably "
                  "should be a TextField or some sublcass of TextField."
    )
    tags = models.TextField(
        blank=True,
        default=_j(whitelists.basic_content_tags),
        help_text="A comma-separated whitelist of HTML tags that are allowed "
                  "in the selected field"
    )
    attributes = models.TextField(
        blank=True,
        default=_j(whitelists.attributes),
        help_text="A comma-separated whitelist of attributes that are "
                  "allowed in the selected field"
    )
    styles = models.TextField(
        blank=True,
        help_text="A comma-separated whitelist of allowed CSS properties "
                  "within a style attribute. NOTE: For this to work, 'style' "
                  "must be in the list of attributes."
    )
    strip = models.BooleanField(
        default=False,
        help_text="Strip disallowed HTML instead of escaping it."
    )
    strip_comments = models.BooleanField(
        default=True,
        help_text="Strip HTML comments."
    )

    def __str__(self):
        return "%s - %s" % (self.content_type, self.field_name)

    class Meta:
        ordering = ['content_type', 'field_name', ]
        unique_together = (('content_type', 'field_name'), )

    @property
    def app_name(self):
        """The name of the App to which this sanitizer is associated"""
        return self.content_type.app_label

    @property
    def model_name(self):
        """The name of the Model to which this sanitizer is associated """
        return self.content_type.model

    def save(self, *args, **kwargs):
        """Checks to see that ``field_name`` is an attribute of the selected
        Model, then registers the signal handler with the appropriate model.
        """
        msg = "The field_name '{0}' does not exist in the model '{1}'".format(
            self.field_name,
            self.content_type.model
        )
        assert self._field_name_in_model(), msg
        super().save(*args, **kwargs)
        _register(sanitize_fields, [self.content_type])

    def _field_name_in_model(self):
        m = self.content_type.model_class()
        return self.field_name in [f.name for f in m._meta.fields]

    def _split(self, text, delimiter=","):
        """Split text by delimiter and and filter out empty values."""
        items = [i.strip() for i in text.split(delimiter)]
        return list(filter(None, items))  # remove blanks

    def get_tags_list(self):
        return self._split(self.tags)

    def get_attributes_list(self):
        return self._split(self.attributes)

    def get_styles_list(self):
        return self._split(self.styles)

    def get_bleach_clean_args(self):
        """Return a dict appropriate for passing into ``bleach.clean``."""
        return {
            'tags': self.get_tags_list(),
            'attributes': self.get_attributes_list(),
            'styles': self.get_styles_list(),
            'strip': self.strip,
            'strip_comments': self.strip_comments
        }


def sanitize_fields(sender, **kwargs):
    """The signal handler for a FieldSanitizer

    * `sender` - the model class
    * `instance` - an instance of the sender

    """
    sender_content_type = ContentType.objects.get_for_model(sender)
    sender_instance = kwargs['instance']

    sanitizers = FieldSanitizer.objects.filter(content_type=sender_content_type)
    for sanitizer in sanitizers:
        if hasattr(sender_instance, sanitizer.field_name):
            field_content = getattr(
                sender_instance,
                sanitizer.field_name
            )
            # Clean with bleach!
            field_content = clean(
                field_content,
                **sanitizer.get_bleach_clean_args()
            )
            setattr(sender_instance, sanitizer.field_name, field_content)


def _clean_class_objects(klass_list):
    """Cleans the content for all classes in the provided list.
    This is done by forcing each instance of the class to
    invoke it's `save` method.

    Returns the total number of objects saved.

    This function is used in the management commands.
    """
    object_count = 0
    for klass in klass_list:
        for obj in klass.objects.all():
            obj.save()
            object_count += 1
    return object_count


class HTMLTagParser(HTMLParser):
    """A Simple parser that stores a set of tags in a document."""

    def add_tag(self, tag):
        if not hasattr(self, "tags"):
            self.tags = set()
        self.tags.add(tag)

    def handle_starttag(self, tag, attrs):
        self.add_tag(tag)


def _get_tags_used_in_content(app_label=None, model=None):
    """
    Use html5lib's parser to get a list of HTML tags used in content
    associated with a FieldSanitizer.

    This can be useful when determining what to include in a whitelist,
    and is used in the `list_html_elements` and
    `list_html_elements_for_model` management commands.

    """
    tag_parser = HTMLTagParser()
    queryset = FieldSanitizer.objects.all()
    if app_label and model:
        queryset = queryset.filter(
            content_type__app_label=app_label,
            content_type__model=model
        )

    for fs in queryset:
        model_class = fs.content_type.model_class()
        for content in model_class.objects.values_list(fs.field_name, flat=True):
            tag_parser.feed(content)

    return sorted(list(tag_parser.tags))


@transaction.atomic
def register_everything():
    """
    This function attempts to register all `FieldSanitizer` instances
    with the `sanitize_fields` callback.

    See: janitor.apps.JanitorConfig.ready()

    """
    try:
        with transaction.atomic():
            _content_type_ids = FieldSanitizer.objects.values_list('content_type')
            _content_type_ids = _content_type_ids.distinct()
            _content_types = [
                ct for ct in ContentType.objects.filter(id__in=_content_type_ids)
            ]
            _register(sanitize_fields, _content_types)
    except IntegrityError:
        transaction.rollback()



"""
Management utility to clean a specific Model that is registered with a
FieldSanitizer.

"""
from django.core.management.base import CommandError, LabelCommand
from janitor.models import FieldSanitizer, _clean_class_objects


class Command(LabelCommand):
    help = "Cleans fields for a specified Model"
    args = "[appname.modelname]"
    label = 'application name.model name'

    def handle_label(self, label, **options):
        try:
            app_label, model = label.lower().split('.')
        except ValueError:
            raise CommandError(
                "Invalid app_label.model_name: {0}".format(label)
            )

        self.stdout.write("\nCleaning {0}.{1}\n".format(app_label, model))

        qs = FieldSanitizer.objects.filter(
            content_type__app_label=app_label,
            content_type__model=model
        )
        try:
            assert qs.count() > 0
        except AssertionError:
            msg = "There are no FieldSanitizers defined for {0}.{1}"
            raise CommandError(msg.format(app_label, model))

        klass_list = [fs.content_type.model_class() for fs in qs]
        object_count = _clean_class_objects(klass_list)
        self.stdout.write("Cleaned {0} Objects\n".format(object_count))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-30 21:36
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='FieldSanitizer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('field_name', models.CharField(help_text='The name of a field in the selected Model. It probably should be a TextField or some sublcass of TextField.', max_length=255)),
                ('tags', models.TextField(blank=True, default='a, abbr, acronym, blockquote, cite, code, dd, del, dfn, dl, dt, em, h1, h2, h3, h4, h5, h6, hr, img, ins, kbd, li, ol, p, pre, q, samp, strong, ul', help_text='A comma-separated whitelist of HTML tags that are allowed in the selected field')),
                ('attributes', models.TextField(blank=True, default='alt, class, href, id, src, title', help_text='A comma-separated whitelist of attributes that are allowed in the selected field')),
                ('styles', models.TextField(blank=True, help_text="A comma-separated whitelist of allowed CSS properties within a style attribute. NOTE: For this to work, 'style' must be in the list of attributes.")),
                ('strip', models.BooleanField(default=False, help_text='Strip disallowed HTML instead of escaping it.')),
                ('strip_comments', models.BooleanField(default=True, help_text='Strip HTML comments.')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'ordering': ['content_type', 'field_name'],
            },
        ),
        migrations.AlterUniqueTogether(
            name='fieldsanitizer',
            unique_together=set([('content_type', 'field_name')]),
        ),
    ]



from django.test import TestCase
from django.contrib.contenttypes.models import ContentType

from janitor import whitelists
from janitor.models import FieldSanitizer


class TestJanitor(TestCase):

    @classmethod
    def setUpTestData(cls):
        # Hack: Use the FieldSanitizer models as the target, so we get a
        # field sanitizer that cleans other field sanitizers' syles field.
        cls.ct = ContentType.objects.get_for_model(FieldSanitizer)
        cls.fs = FieldSanitizer.objects.create(
            content_type=cls.ct,
            field_name="styles",
            tags="h1, p, a",
            attributes="href",
            styles="",
            strip=True,
            strip_comments=True,
        )

        # some content that we'll try to clean
        cls.sample_content = ("""
<h1 id="a"><a href="/">Heading</a></h1>
<p style="color:blue;" class="foo">Blue <strong>text</strong>
<script></script></p>
        """)

        # Cleaned results to verify that everything's working
        cls.cleaned_content = ("""
<h1><a href="/">Heading</a></h1>
<p>Blue text
</p>
        """)

    def setUp(self):
        self.obj = FieldSanitizer.objects.create(
            content_type=self.ct,
            field_name="field_name",
            styles=self.sample_content
        )

    def tearDown(self):
        self.obj.delete()

    def test__str__(self):
        """Tests the FieldSanitizer.__str__ method."""
        self.assertEqual(self.fs.__str__(), "field sanitizer - styles")

    def test_app_name(self):
        self.assertEqual(self.fs.app_name, "janitor")

    def test_model_name(self):
        self.assertEqual(self.fs.model_name, "fieldsanitizer")

    def test__field_name_in_model(self):
        self.assertTrue(self.fs._field_name_in_model())

    def test__split(self):
        self.assertEqual(self.fs._split("a,b,c,"), ['a', 'b', 'c'])

    def test_get_tags_list(self):
        self.assertEqual(self.fs.get_tags_list(), ['h1', 'p', 'a'])

    def test_get_attributes_list(self):
        self.assertEqual(self.fs.get_attributes_list(), ['href'])

    def test_get_styles_list(self):
        self.assertEqual(self.fs.get_styles_list(), [])

    def test_default_clean(self):
        """Creates an instance of the test model with some sample content,
        then verifies that it gets cleaned upon saving.
        """
        self.assertEqual(self.obj.styles, self.cleaned_content)

    def test_get_bleach_clean_args(self):
        actual = self.fs.get_bleach_clean_args()
        expected = {
            'tags': ['h1', 'p', 'a'],
            'attributes': ['href'],
            'styles': [],
            'strip': True,
            'strip_comments': True,
        }
        self.assertDictEqual(actual, expected)

    def test_strip_content(self):
        """Adds an HTML comment to the class's sample content, then verifies
        that it gets removed.
        """
        self.obj.styles = "<!-- Hello! -->" + self.sample_content
        self.obj.save()
        self.assertEqual(self.obj.styles, self.cleaned_content)

    def test_default_object_data(self):
        """Tests that FieldSanitizer's default tags, attributes, and clean_args
        are in place.
        """
        fs = FieldSanitizer()
        self.assertEqual(fs.get_tags_list(), whitelists.basic_content_tags)
        self.assertEqual(fs.get_attributes_list(), whitelists.attributes)

        expected_args = {
            'styles': [],
            'attributes': [u'alt', u'class', u'href', u'id', u'src', u'title'],
            'strip': False,
            'strip_comments': True,
            'tags': [
                u'a', u'abbr', u'acronym', u'blockquote', u'cite', u'code',
                u'dd', u'del', u'dfn', u'dl', u'dt', u'em', u'h1', u'h2',
                u'h3', u'h4', u'h5', u'h6', u'hr', u'img', u'ins', u'kbd',
                u'li', u'ol', u'p', u'pre', u'q', u'samp', u'strong', u'ul'
            ]
        }
        self.assertEqual(fs.get_bleach_clean_args(), expected_args)



"""
Management utility to clean a specific Model that is registered with a
FieldSanitizer.
"""
from django.core.management.base import CommandError, LabelCommand
from janitor.models import _get_tags_used_in_content


class Command(LabelCommand):
    help = "Lists HTML elements used in a Model with a related FieldSanitizer"
    args = "[appname.modelname]"
    label = 'application name.model name'

    def handle_label(self, label, **options):
        try:
            app_label, model = label.lower().split('.')
        except ValueError:
            raise CommandError("Invalid app_label.model_name: {0}".format(label))

        tag_list = _get_tags_used_in_content(app_label, model)
        tag_list = ", ".join(sorted(tag_list))
        msg = "\nTags used in {}.{}: {}\n".format(app_label, model, tag_list)
        self.stdout.write(msg)



from django.test import TestCase

from janitor.models import FieldSanitizer
from janitor.forms import FieldSanitizerAdminForm


class TestFieldSanitizerAdminForm(TestCase):
    def setUp(self):
        self.form = FieldSanitizerAdminForm()

    def test_form(self):
        self.assertEqual(self.form.Meta.model, FieldSanitizer)
        expected_fields = [
            'attributes',
            'content_type',
            'field_name',
            'strip',
            'strip_comments',
            'styles',
            'tags',
        ]
        self.assertEqual(sorted(self.form.fields), expected_fields)

    def test_expected_output(self):
        """Verify that the ContentTypeChoiceFieldWithAppLabel field displays
        things as expected."""
        # Assuming the auth/user content type is installed
        self.assertIn("auth/user", str(self.form))
